import math
import numpy as np
from sklearn.preprocessing import KBinsDiscretizer


class DistributionBuilder:
    """
    Class that turns a list of properties into a normalized distribution
    feature_type: [categorical (cat), categorical_multi (cat_m), continuous (cont)]
    rank_aware: should the values be discounted based on their rank in the recommendation
    bins: optional - into how many bins should continuous values be divided
    """

    def __init__(self, feature_type, rank_aware, **kwargs):
        self.feature_type = feature_type
        self.rank_aware = rank_aware

        if self.feature_type == 'cont':
            bins = kwargs.get('bins', 10)
            self.bins_discretizer = KBinsDiscretizer(
                encode='ordinal', n_bins=bins, strategy='uniform', subsample=None)

    def build_distribution(self, x):
        if self.feature_type == 'cat':
            return self.categorical(x)
        elif self.feature_type == 'cont':
            return self.continuous(x)
        elif self.feature_type == 'cat_m':
            return self.categorical_multi(x)

    @staticmethod
    def harmonic_number(n):
        """Returns an approximate value of n-th harmonic number.
        http://en.wikipedia.org/wiki/Harmonic_number
        """
        # Euler-Mascheroni constant
        gamma = 0.57721566490153286060651209008240243104215933593992
        return gamma + math.log(n) + 0.5 / n - 1. / (12 * n ** 2) + 1. / (120 * n ** 4)

    def categorical(self, x):
        """"
        Parameters
        ----------
        x : List of properties, where the first entry refers to the first article, the second entry to the second, etc.
        r : Boolean. Should the distribution be discounted or not
        Returns
        -------
        Dictionary where every entry refers to the presence of that property in the distribution.

        """
        n = len(x)
        sum_one_over_ranks = self.harmonic_number(n)
        distribution = {}
        count = 0
        for _, item in enumerate(x):
            count += 1
            feature_freq = distribution.get(item, 0.)
            distribution[item] = feature_freq + 1 * 1 / count / \
                                 sum_one_over_ranks if self.rank_aware else feature_freq + 1 * 1 / n
        return distribution

    def categorical_multi(self, x):
        """"
        Build distributions where the relevant feature can have multiple values. For example, multiple people can be
        mentioned in an article, or an article can have multiple topics/categories assigned.
        Parameters
        ----------
        x : List of properties, where the first entry refers to the first article, the second entry to the second, etc.
        Returns
        -------
        Dictionary where every entry refers to the presence of that property in the distribution.

        """
        n = len(x)
        sum_one_over_ranks = self.harmonic_number(n)
        distribution = {}
        for i, item in enumerate(x):
            for j, entry in enumerate(item):
                rank = i + 1
                feature_freq = distribution.get(entry, 0.)
                distribution[entry] = feature_freq + 1 * 1 / rank / \
                                     sum_one_over_ranks if self.rank_aware else feature_freq + 1 * 1 / n
        # normalizing the distribution is a bit harder when it's unknown how many entities should be accounted for
        if len(distribution) > 0:
            factor = 1.0 / sum(distribution.values())
            for k in distribution:
                distribution[k] = distribution[k] * factor
            return distribution
        else:
            return None

    def continuous(self, x):
        """"
        List of continuous values. Since the divergence-based metric is essentially categorical, these values are first
        binned. This means that we lose information about ordering, and could use improvement in the future.
        Parameters
        ----------
        x : List of properties, where the first entry refers to the first article, the second entry to the second, etc.
        Returns
        -------
        Dictionary where every entry refers to the presence of that property in the distribution.

        """
        n = len(x)
        sum_one_over_ranks = self.harmonic_number(n)
        arr_binned = self.bins_discretizer.transform(np.array(x).reshape(-1, 1))
        distribution = {}
        if self.rank_aware:
            for bin in list(range(self.bins_discretizer.n_bins)):
                for i, ele in enumerate(arr_binned[:, 0]):
                    if ele == bin:
                        rank = i + 1
                        bin_freq = distribution.get(bin, 0.)
                        distribution[bin] = bin_freq + 1 * 1 / rank / sum_one_over_ranks
        else:
            for bin in list(range(self.bins_discretizer.n_bins)):
                distribution[bin] = round(np.count_nonzero(
                    arr_binned == bin) / arr_binned.shape[0], 3)
        return distribution



import pandas as pd
import ast
import numpy as np

from nltk.sentiment.vader import SentimentIntensityAnalyzer
import nltk
nltk.download('vader_lexicon')

articles = pd.DataFrame()
behaviors = pd.DataFrame()
predictions = pd.DataFrame()


"""
Article preprocessing. This includes:
- Sentiment analysis
- Adding an extra column that contains the people mentioned in the title entities and subtitle entities columns
"""

# Initialize the VADER sentiment intensity analyzer
sid = SentimentIntensityAnalyzer()


# Define a function to calculate sentiment scores
def get_sentiment_score(text):
    sentiment = sid.polarity_scores(text)
    return abs(sentiment['compound'])


# Code for extracting the people mentioned in the title and subtitle
def extract_people(row):
    try:
        entries = ast.literal_eval(row)
        labels = [entry['Label'] for entry in entries if entry['Type'] == 'P']
        return labels
    except ValueError:
        return []


def process_articles(location):
    global articles
    articles = pd.read_csv(location, sep='\t',
                           names=['article_id', 'category', 'subcategory', 'title', 'subtitle', 'url', 'entities_title',
                                  'entities_subtitle'], index_col=0)

    # Apply the function to the 'text' column to create a new 'sentiment_score' column
    articles['absolute_sentiment_score'] = articles['title'].apply(get_sentiment_score)

    # Apply function to the column and create new column with the extracted labels
    articles['persons'] = articles['entities_title'].apply(extract_people) + articles['entities_subtitle'].apply(
        extract_people)
    return articles


def make_list(lst, feature):
    """
    Function that helps extract the relevant features for a recommendation. Returns for each impr_index in the list the
    corresponding feature, in the same order as in the list.

    Example for feature 'category':
        articles: [{'impr_index': 'N1', 'category': 'news'},{'impr_index': 'N2', 'category': 'sport'},{'impr_index': 'N3', 'category': 'entertainment'}]
        lst : ['N2','N3','N1']
        output: ['sport','entertainment','news']
    """
    try:
        result = list(articles.loc[lst][feature])
        return result
    except KeyError:
        return []



"""
Preparing the prediction files. This means:
- Filters the predictions to only include the ones that are also in the behaviors file (in case of sampling)
- Adding an extra column that has the articles in the impression ordered by their predicted rank
- Adding an extra column for the date of the prediction
"""


def order(row, recommendation_length):
    prediction = row['pred_rank']
    behavior_row = behaviors.loc[row['impr_index']]
    # match the predicted ordering from pred_rank with the right article id as it is specified in
    # the behavior file impression
    ordered = [behavior_row.impression[prediction[i] - 1] for i in range(min(len(prediction), recommendation_length))]
    return ordered


def process_recommendations(location, sample_size):
    pred = pd.read_json(location)

    columns = list(pred.columns)
    not_algorithms = ['impr_index', 'userid', 'date', 'history']
    algorithms = [ele for ele in columns if ele not in not_algorithms]

    if sample_size > 0:
        unique_users = pred.userid.unique()
        selected_users = np.random.choice(unique_users, size=sample_size, replace=False)
        pred = pred[pred['userid'].isin(selected_users)]
    return algorithms, pred



import numpy as np

from RADio.divergence import Divergence
from RADio.distributions import DistributionBuilder


class DiversityMetric:
    """
    Base class for the diversity metrics. Can be configured to reflect the type of diversity necessary in your
    application.
    feature_type: [categorical (cat), categorical multi (cat_m), continuous (cont)]
    rank_aware_recommendation: boolean; should the values in the recommendation be discounted based on position
    rank_aware_context: boolean; should the values in the context be discounted based on position
    bins: int; optional, number of bins to use for discretization
    context_type: [dynamic, static]; for efficiency; should the context distribution be calculated every time, or is it
                    always the same?
    metric: [JSD, KL]; to use the Jensen-Shannon Divergence or Kullback-Leibler Divergence
    """

    def __init__(self, **kwargs):
        self.feature_type = kwargs.get('feature_type', 'cat')
        self.rank_aware_recommendation = kwargs.get('rank_aware_recommendation', True)
        self.rank_aware_context = kwargs.get('rank_aware_context', True)
        self.bins = kwargs.get('bins', 10)
        self.context_type = kwargs.get('context', 'dynamic')

        self.divergence = Divergence(metric=kwargs.get('metric', 'JSD'))

        self.recommendation_builder = DistributionBuilder(self.feature_type, self.rank_aware_recommendation)
        self.context_builder = DistributionBuilder(self.feature_type, self.rank_aware_context)

        self.Q_static = {}

    def compute(self, recommendation, context):

        if self.context_type == 'dynamic':
            Q = self.context_builder.build_distribution(context)
        else:
            if not self.Q_static:
                if self.feature_type == 'cont':
                    self.context_builder.bins_discretizer.fit(np.array(context).reshape(-1, 1))
                    self.recommendation_builder.bins_discretizer.fit(np.array(context).reshape(-1, 1))
                self.Q_static = self.context_builder.build_distribution(context)
            Q = self.Q_static

        P = self.recommendation_builder.build_distribution(recommendation)

        if P and Q:
            return self.divergence.compute(P, Q)
        else:
            return None



import math
from numpy.linalg import norm
from scipy.stats import entropy


class Divergence:
    """
    Class that calculates the divergence between two distributions P and Q.
    Assumes two dictionaries with the same keys as input
    """

    def __init__(self, metric='JSD'):
        self.metric = metric

    @staticmethod
    def opt_merge_max_mappings(a, b):
        """ Merges two dictionaries based on the largest value in a given mapping.
        Parameters
        ----------
        a : Dict[Any, Comparable]
        b : Dict[Any, Comparable]
        Returns
        -------
        Dict[Any, Comparable]
            The merged dictionary
        """
        merged, other = (a, b) if len(a) > len(
            b) else (b, a)
        merged = dict(merged)
        for key in other:
            if key not in merged or other[key] > merged[key]:
                merged[key] = other[key]
        return merged

    def compute(self, s, q, alpha=0.001):
        """
        KL (p || q), the lower the more similar the distributions are.
        alpha is not really a tuning parameter, it's just there to make the
        computation more numerically stable.
        """
        try:
            assert 0.99 <= sum(s.values()) <= 1.01
            assert 0.99 <= sum(q.values()) <= 1.01
        except AssertionError:
            print("Input not normalized distributions with matching keys")
            print(sum(s.values()))
            print(sum(q.values()))
            pass

        ss = []
        qq = []
        merged_dic = self.opt_merge_max_mappings(q, s)
        for key in sorted(merged_dic.keys()):
            s_score = s.get(key, 0.)
            q_score = q.get(key, 0.)
            qq.append((1 - alpha) * s_score + alpha * q_score)
            ss.append((1 - alpha) * q_score + alpha * s_score)
        if self.metric == 'JSD':
            return self.JSD(ss, qq)
        elif self.metric == 'KL':
            return entropy(ss, qq, base=2)

    @staticmethod
    def JSD(P, Q):
        """ Compute J-S divergence.
            Parameters
            ----------
            P : Dictionary. distribution of pool.
            Q : Dictionary. distribution of recommendation.
            Returns
            -------
            JS divergence value.

        """
        _P = P / norm(P, ord=1)
        _Q = Q / norm(Q, ord=1)
        _M = 0.5 * (_P + _Q)
        try:
            jsd_root = math.sqrt(
                abs(0.5 * (entropy(_P, _M, base=2) + entropy(_Q, _M, base=2))))
        except ZeroDivisionError:
            jsd_root = None
        return jsd_root



import plotly.figure_factory as ff
import plotly.express as px
import seaborn as sns
import pandas as pd


def distplot_per_user(df, algorithms, metric):
    data = []
    for algorithm in algorithms:
        user_df =df.groupby('userid')[algorithm].mean().reset_index()
        user_df = user_df[user_df[algorithm].notna()]
        data.append(user_df[algorithm].tolist())

    group_labels = algorithms
    fig = ff.create_distplot(data, group_labels, bin_size=0.01, show_hist=False)
    fig.update_layout(title='Dist Plot of average values for ' + metric + ' per user')
    fig.show()


def distplot_all(df, algorithms, metric):
    data = []
    df1 = df[df[metric].notna()]
    for algorithm in algorithms:
        data.append(df1[df1['algorithm'] == algorithm][metric])

    fig = ff.create_distplot(data, algorithms, bin_size=.01, show_hist=False)
    fig.show()


def lineplot_over_time(df, metric):
    df['date'] = pd.to_datetime(df['date'])
    df['hour'] = df['date'].dt.hour
    df['per_hour'] = df.date.dt.round(freq='1h')

    hourly_avg = df.groupby(['per_hour', 'algorithm'])[metric].mean().reset_index()
    fig = px.line(hourly_avg, x='per_hour', y=metric, color="algorithm")
    fig.show()


def lineplot(df, metric):
    df['date'] = pd.to_datetime(df['date'])
    df['per_hour'] = df.date.dt.round(freq='1h')
    result = df.drop(columns=['userid', 'date'])
    hourly_avg = result.groupby('per_hour').mean().reset_index()
    df_melted = hourly_avg.melt(id_vars='per_hour', var_name='algorithm', value_name='average_value')
    fig = px.line(df_melted, x='per_hour', y='average_value', color='algorithm', title='Average Values for ' + metric +' Over Time')
    fig.show()

def seaborn_per_hour(df, metric):
    df['date'] = pd.to_datetime(df['date'])
    df['hour'] = df['date'].dt.hour
    df['per_hour'] = df.date.dt.round(freq='1h')

    sns.set_theme(rc={'figure.figsize': (15, 6)})
    sns.lineplot(x="per_hour", y=metric,
                 hue="algorithm",
                 data=df)


def visualize(df, metric, algorithms):
    result = df[['userid', 'date', metric]]
    for i, name in enumerate(algorithms):
        result[name] = result[metric].apply(lambda x: x[i])
    result = result.drop(columns=[metric])
    print(result.describe())

    distplot_per_user(result, algorithms, metric)
    lineplot(result, metric)
